#!/usr/bin/perl

use strict;

#TK Dependencies
use Tk;
use Tk::Dialog;
use Tk::DialogBox;
use Tk::LabEntry;
use Tk::Text;
use Tk::FileSelect;
use Tk::BrowseEntry;
use Tk::Entry;
use Tk::Scrollbar;
use Tk::Menubutton;
use Tk::Widget;
use Tk::Photo;
use Tk::Font;
use Tk::HList;
use Tk::Checkbutton;
use Tk::Radiobutton;
use Tk::FileSelect;
use Cwd;

#network dependices
use IO::Socket;
use Time::HiRes qw(sleep);
use Data::Dumper;
use threads;
use Thread;


my $tk_version = $Tk::VERSION;

$tk_version =~ /(\d+)..*/;
$tk_version = $1;

my $ssh_thread;
my $show_only = "All";
my $source_ip;
my $text;
my $mangement_password;
my $refresh_count = 0;
my $data_hash;
my $display_hash;
my $entry;
my $menubar;
my $file_menu;
my $edit_menu;
my $view_menu;
my $sort_order_menu;
my $sort_direction_menu;
my $help_menu;
my $config;
my $initial_ssh_session = "no";
my $sort_direction = "ascending";
my $sort_header = "username";
my $os;
my $connection;
my $hlist;
my $number_of_connections;
my $refresh_index = 0;
my $win_ssh_thread;
my $config_fields;
my $server_popup;

if ($ENV{'windir'}){
	$os = "windows";
	#use Win32::Process;
} else {
	$os = "other";
}

my $refresh;
my $preferences = get_preferences();

my @server_array = sort keys(%{$$preferences{server}});
my $server = $server_array[0];
	
#lines needed for compilation under windows
#perl2exe_exclude "Win32Util.pm"
#perl2exe_include "Tk/arrowdownwin.xbm"
#perl2exe_include "utf8.pm"
#perl2exe_include "unicore/lib/gc_sc/Word.pl";
#perl2exe_include "unicore/lib/gc_sc/Digit.pl";
#perl2exe_include "unicore/lib/gc_sc/SpacePer.pl";
#perl2exe_include "unicore/To/Lower.pl";
#perl2exe_include "unicore/lib/gc_sc/Cntrl.pl";
#perl2exe_include "unicore/lib/gc_sc/ASCII.pl";
#perl2exe_include "unicore/To/Fold.pl";



#create main screen
#Tk::CmdLine::SetArguments(qw(-geometry 741x374 ));

my $main = MainWindow->new(-background => "cadetblue",
			   			 -title => "OpenVPN-Control");
			   			 
my $icon = $main->Photo(-file => 'openvpn-control.gif');
 
$main->iconimage($icon);

#install signal hangler to quit out of SSH sessions window and do other cleanup at close
$main->protocol('WM_DELETE_WINDOW',\&exit_handler);			

$menubar = $main-> Menu();
$main -> configure(-menu => $menubar);                          


$file_menu = $menubar -> cascade(-label=>"File", -underline=>0, -tearoff => 0);
$file_menu -> command(-label => "Configure OpenVPN Control", -underline=>0, 
		-command=>\&configure_control );
$file_menu -> command(-label => "Exit", -underline=>0, 
		-command=>\&exit_handler  );


$edit_menu = $menubar -> cascade(-label=>"Edit", -underline=>0, -tearoff => 0);
$edit_menu -> command(-label => "Copy", -underline=>0, 
		-command=>\&Copy_to_Clipboard );
$edit_menu -> separator();
$edit_menu -> command(-label => "Select All", -underline=>0, 
		-command=>\&Select_All );
$edit_menu -> command(-label => "Clear Selection", -underline=>0, 
		-command=>\&Select_Clear );		

$view_menu = $menubar -> cascade(-label=>"View", -underline=>0, -tearoff => 0);

$sort_direction_menu = $view_menu -> cascade(-label=>"Sort Direction", -underline=>0, -tearoff => 0);
$sort_direction_menu -> radiobutton(-label => "Sort Ascending", 
 							       -underline=>0,
 							       -variable => \$sort_direction,
 							       -value => "ascending",
								   -command => [ sub { get_login_info(); }] );

$view_menu->separator();
							   
$sort_direction_menu -> radiobutton(-label => "Sort Descending", 
								   -underline=>0,
								   -variable => \$sort_direction,
								   -value => "descending",
								   -command => [ sub { get_login_info(); }] );

$sort_order_menu = $view_menu -> cascade(-label=>"Sort Order", -underline=>0, -tearoff => 0);

my @header_list = ("Server", "Source IP", "User Name", "Assigned Address", "Transfered", "Start Time", "Last Active");
my @header_value = ("server", "source_ip", "username", "assign_ip", "transfered", "start_time", "last_active");
my $value = 0;
foreach (@header_list){
	$sort_order_menu -> radiobutton(-label => "$_", 
 							       -underline=>0,
 							       -variable => \$sort_header,
 							       -value => "$header_value[$value]",
								   -command => [ sub { get_login_info(); }] );
	$value++;
}								   


$view_menu->separator();

$view_menu -> command(-label => "Refresh", -underline=>0, 
					-command => \&get_login_info );


my	$system_menu = $menubar -> cascade(-label=>"Servers", -underline=>0, -tearoff => 0);
	$system_menu -> radiobutton(-label => "All", 
							  -underline=>0,
							  -variable => \$show_only,
							  -value => "All",
					 	  	  -command=> [ sub { $preferences = get_preferences();
					 	  	  					get_login_info(); } ]);	

	$system_menu->separator();
	
	foreach (@server_array){
		$system_menu -> radiobutton(-label => "$_", 
 							       -underline=>0,
 							       -variable => \$show_only,
 							       -value => "$_",
								   -command=> [ sub { show_only_one_server($show_only);}] );
	}


$help_menu = $menubar -> cascade(-label=>"Help", -underline=>0, -tearoff => 0);
$help_menu -> command(-label => "About", -underline=>0, 
					 -command=> \&about );
					
my $connections_label = 	$main->Label(-text => "Number of Connections: ",
									-background => 'cadetblue',
	 				     			)->place(-x=>'2', -y=>'3');
my $connections;
#two differenct version in an attempt to work across multiple TK versions
#was working with readonly on Windows and OSX, but not Debian for some reason.
if ($tk_version > 801){
	$connections = $main->Entry (-background => "white",
							  -textvariable => \$number_of_connections,
							  -width => '4',
							  -state => 'readonly',
							  )->place(-x=>'150', -y=>'3');
} else {
	$connections = $main->Entry (-background => "white",
							  -textvariable => \$number_of_connections,
							  -width => '4',
							  -state => 'disabled',
							  -background => 'gray',
							  )->place(-x=>'150', -y=>'3');
}
							  							  
#create buttons
my $log_off = $main->Button( -text   => "Force Logoff",
               -command => \&force_logoff )->pack;

 
$hlist = $main->Scrolled ('HList',
			-columns=>7,
			-width=>127,
			-scrollbars => 'e',
			-height=>19,
			-header=>1,
			-selectmode=>'multiple');
			
#create header
my $count = 0;
foreach (@header_list) {
	$hlist->header('create', $count, -text => $_, 
								    -headerbackground =>  'gray');
	$count++;
}
$count = 0;

$hlist->columnWidth(0, -char, 20);
$hlist->columnWidth(1, -char, 23);
$hlist->columnWidth(2, -char, 15);
$hlist->columnWidth(3, -char, 20);
$hlist->columnWidth(4, -char, 16);
$hlist->columnWidth(5, -char, 17);
$hlist->columnWidth(6, -char, 17);

$hlist->pack(-side => 'left', -expand=>'yes', -fill=>'both');

$main->update();
sleep(1);
refresh();

#exit;

sub Select_All {
	
	my $count = $number_of_connections - 1;
	
	$hlist->selectionSet(0,$count);
	return;
	
}

sub Select_Clear {
	
	my $count = $number_of_connections - 1;

	$hlist->selectionClear(0,$count);
	
	get_login_info();
	
	return
}

sub show_only_one_server {
	
	my $server = shift;
	foreach (@server_array){
		$$preferences{server}{$_}{enabled} = 0;
	}	
	
	$$preferences{server}{$server}{enabled} = 1;
	get_login_info();
	
	return();
}

sub refresh {
	
	get_login_info();

	while(){

		$main->update();
		sleep (.25);
		#refesh display every 60 seconds
		if ($refresh_count > $refresh){
			get_login_info();
			$refresh_count = 0;
		}

		$refresh_count++;
			
	}

	
}

sub Copy_to_Clipboard {
	
	my @index = $hlist->info('selection');
	my $text;
	my $record;
	my $header_value;
	my $value;
	my $line;
	
	$main->clipboardClear;
	
	foreach $record (@index){
		undef($line);
		foreach $header_value (@header_value){
			 $value = $$display_hash{$record}{$header_value};
			 $value =~ s/^\s+//;
			 $value =~ s/\s+$//;
			 
			 $line .= "\"$value\"\t";
		}
		
		$line =~ s/\t$//; 
		$text .= $line . "\n";
	}
		
	$main->clipboardAppend('--',$text);
	
	$main->SelectionClear;
	$main->SelectionHandle( sub {$text} );
	$main->SelectionOwn;
	
}

sub force_logoff {

	my @index = $hlist->info('selection');
	my $text = "";
	my $username;
	my $sourceip;
	
	foreach (@index){
		$username = $$display_hash{$_}{username};
		$text .= "${username}::"; 
	}
	
	#clean up of spaces
	$text =~ s/\s//g;
	$text =~ s/::/\n/g;
	$text =~ s/,$//;

	if ($$display_hash{0}{source_ip}){
		confirm("You are about to log off:\n$text\nAre you sure?");
		foreach (@index){
			$sourceip =  $$display_hash{$_}{source_ip};
			$sourceip =~ s/^\s+//;
			$sourceip =~ s/\s+$//;
			
			$server =  $$display_hash{$_}{server};
			$server =~ s/^\s+//;
			$server =~ s/\s+$//;
			openvpn_management("kill $sourceip", $server);
		}
		
		Info_Message ("Connection for:\n$text\nhas been terminated!!");
		get_login_info();
	}
	
	return;

}

sub get_login_info {
	
	my $server;
	my $index;
	my $count = 0;

	$display_hash = undef;
	$display_hash = display_hash_generator();
	
	#print "$refresh - $refresh_count\n";

	$entry  = undef;
	$hlist->delete('all');

	foreach $index (sort keys %{$display_hash}){
		$entry = $hlist->add($index, -at => $index);
		$hlist->itemCreate($entry, 0, -itemtype=>'text', -text=>"$$display_hash{$index}{server}");
		$hlist->itemCreate($entry, 1, -itemtype=>'text', -text=>"$$display_hash{$index}{source_ip}");
		$hlist->itemCreate($entry, 2, -itemtype=>'text', -text=>"$$display_hash{$index}{username}");
		$hlist->itemCreate($entry, 3, -itemtype=>'text', -text=>"$$display_hash{$index}{assign_ip}");
		$hlist->itemCreate($entry, 4, -itemtype=>'text', -text=>"$$display_hash{$index}{transfered}");
		$hlist->itemCreate($entry, 5, -itemtype=>'text', -text=>"$$display_hash{$index}{start_time}");
		$hlist->itemCreate($entry, 6, -itemtype=>'text', -text=>"$$display_hash{$index}{last_active}");
		$count++;
			
	}
	
	$hlist->update;
	
	$number_of_connections = $count;

	
	return($display_hash);
	
}

sub display_hash_generator {

	my $index;
	my $entry;
	my $sorted_hash;
	my $sort_key;
	
	my $results = "";
	
	foreach $server (@server_array){
		if ($$preferences{server}{$server}{enabled} == 1){
	 		$results .= openvpn_management("status 1", $server); 
		} 
	}
	
	$data_hash = parse_status ($results);

	#create sorted_hash
	foreach $source_ip (sort keys %{$data_hash}){
		$$sorted_hash{$$data_hash{$source_ip}{$sort_header}}{$source_ip}{username} = $$data_hash{$source_ip}{username};
	 	$$sorted_hash{$$data_hash{$source_ip}{$sort_header}}{$source_ip}{start_time} = $$data_hash{$source_ip}{start_time};
	 	$$sorted_hash{$$data_hash{$source_ip}{$sort_header}}{$source_ip}{last_active} = $$data_hash{$source_ip}{last_active};
	 	$$sorted_hash{$$data_hash{$source_ip}{$sort_header}}{$source_ip}{assign_ip} = $$data_hash{$source_ip}{assign_ip};
	 	$$sorted_hash{$$data_hash{$source_ip}{$sort_header}}{$source_ip}{transfered} = $$data_hash{$source_ip}{transfered};
	 	$$sorted_hash{$$data_hash{$source_ip}{$sort_header}}{$source_ip}{server} = $$data_hash{$source_ip}{server};	
	}		

	$count = 0;
	
	if ($sort_direction eq "descending"){
		if ($sort_header eq "transfered"){
			foreach $sort_key (sort {$b <=> $a} keys %{$sorted_hash}){
				foreach $source_ip (sort keys %{$$sorted_hash{$sort_key}}){
					if ( $$data_hash{$source_ip}{refresh_count} == $refresh_index ){
						$display_hash = create_display_hash($source_ip, $count);
						$count++;
					
					}
     		 	}
			}			
			
		} else {
			foreach $sort_key (sort {$b cmp $a} keys %{$sorted_hash}){
				foreach $source_ip (sort keys %{$$sorted_hash{$sort_key}}){
					if ( $$data_hash{$source_ip}{refresh_count} == $refresh_index ){
						$display_hash = create_display_hash($source_ip, $count);
						$count++;
					
					}
     		 	}
			}
		}
	} else {
		if ($sort_header eq "transfered"){
			foreach $sort_key (sort {$a <=> $b} keys %{$sorted_hash}){
				foreach $source_ip (sort keys %{$$sorted_hash{$sort_key}}){
					if ( $$data_hash{$source_ip}{refresh_count} == $refresh_index ){
					
						$display_hash = create_display_hash($source_ip, $count);
						$count++;
					
					}
     		 	}
			}			
			
		} else {
			foreach $sort_key (sort {$a cmp $b } keys %{$sorted_hash}){
				foreach $source_ip (sort keys %{$$sorted_hash{$sort_key}}){
					if ( $$data_hash{$source_ip}{refresh_count} == $refresh_index ){
					
						$display_hash = create_display_hash($source_ip, $count);
						$count++;
					
					}
     		 	}
			}
		}
	}

	return ($display_hash);
}

sub create_display_hash{

	my $source_ip = shift;
	my $count = shift;

	$$display_hash{$count}{source_ip} = sprintf ("%22s", $source_ip);
	$$display_hash{$count}{server} = sprintf ("%12s", $$data_hash{$source_ip}{server});
	$$display_hash{$count}{start_time} = sprintf ("%12s", $$data_hash{$source_ip}{start_time});
	$$display_hash{$count}{assign_ip}  = sprintf ("%15s", $$data_hash{$source_ip}{assign_ip});
	$$display_hash{$count}{username} = sprintf ("%15s", $$data_hash{$source_ip}{username});
	$$display_hash{$count}{last_active} = sprintf ("%12s",$$data_hash{$source_ip}{last_active});
		
	$$data_hash{$source_ip}{transfered} = $$data_hash{$source_ip}{transfered} / 1048576;
		
	$$display_hash{$count}{transfered} = sprintf ("%.2f MB",$$data_hash{$source_ip}{transfered});
	$$display_hash{$count}{transfered} = sprintf ("%10s",$$display_hash{$count}{transfered} );
	
	return($display_hash);
	
}

sub openvpn_management {

	my $command = shift;
	my $server = shift;
	my $results;
	my $status = "ok";
	
	my $port;
	
	if ($$preferences{server}{$server}{ssh_tunnel} == 1){
		$port = $$preferences{server}{$server}{ssh_port_forward};
	} else {
		$port = $$preferences{server}{$server}{port};
	
	}
	
	if ($$preferences{server}{$server}{ssh_tunnel} == 1){
	 	$connection = IO::Socket::INET->new( PeerAddr => $$preferences{server}{$server}{server_ip},
					     PeerPort => $port,
					     Proto => "tcp",
					     Timeout => "10",
					     Type => SOCK_STREAM) or ssh_tunnel($server);
	}
	
	$connection = IO::Socket::INET->new( PeerAddr => $$preferences{server}{$server}{server_ip},
					     PeerPort => $port,
					     Proto => "tcp",
					     Timeout => "10",
					     Type => SOCK_STREAM) or $status = "error";
					     
	if (($status eq "error") | (!($connection))){
		Error_Message("Could not establish connection with $server via $port");
		$$preferences{server}{$server}{enabled} = 0;
		return;
	}
			     
	if ($$mangement_password{$server}){
		print $connection "$$mangement_password{$server}\n\n\n\n";
		confirm_password($connection);
		print $connection "$command\n\n\n\n";
	
	} else {
		print $connection "$command\n\n\n";
		
	}
	
	while (<$connection>){
	
		$results .= "$server:" . $_;

		if (($_ =~ /END/) | ($_ =~ /^SUCCESS/)){
			last;
		}
		
		if (($_ =~ /PASSWORD/) & (!($$mangement_password{$server}))){
			
			$$mangement_password{$server} = get_password($server);
			$results = openvpn_management($command, $server);
			last;
		
		} 
		
		
		
	}
	
	print $connection "quit\n";
	
	close ($connection);
		     
	return ($results);

}	


sub confirm_password {

	my $connection = shift;
	my $accepted = "false";
	
	while (<$connection>){
		
		if ($_ =~ /password is correct/){
			$accepted = "true";
			last;
		}
	}

	if ($accepted eq "false"){
		$$mangement_password{$server} = get_password($server);
		get_login_info();
		return;
	}
	
	return;

}


sub parse_status {

	my $source_ip;
	my $time;
	my $raw_data = shift;
	my $raw_data_array;
	my $section = "Client List";
	
	$refresh_index++;
				
	@$raw_data_array = split(/\n/, $raw_data);
	foreach (@$raw_data_array){
		
		#parse routing section
		if ($_ =~ /^.*:([^,]+),([^,]+),([^,]+),([^,]+),\w\w\w\s(\w\w\w\s+\d+)\s+(\d+:\d+):\d+\s\d\d\d\d/){



				$time = fix_time($6);
			
				$$data_hash{$2}{start_time} = "$5, $time";
				$$data_hash{$2}{transfered} = $3 + $4;
				$$data_hash{$2}{refresh_count} = $refresh_index;

		} elsif ($_ =~ /^([^:]*):([^,]+),([^,]+),([^,]+),\w\w\w\s(\w\w\w\s+\d+)\s+(\d+:\d+):\d+\s\d\d\d\d/){

				$$data_hash{$4}{username} = $3;
				$$data_hash{$4}{assign_ip} = $2;
				$$data_hash{$4}{server} = $1;
			
				$time = fix_time($6);
			
				$$data_hash{$4}{last_active} = "$5, $time";			

				$$data_hash{$4} =~ s/^\s+//;
		
		}

	}
	close (FH);
	
	return ($data_hash);

}

sub get_preferences {

	my $conf_file = "openvpn_control.conf";
	my $preferences;
	
	open (FH, "< $conf_file");
	
	while (<FH>){
	
		if (!($_ =~ /#/)){
			if ($_ =~ /^(.*):(.*)=(.*)$/){
				
				$$preferences{server}{$1}{$2} = $3;
			
			} elsif ($_ =~ /^(.*)=(.*)$/){
				$$preferences{$1} = $2;
					
			} 
		}
	
	} 
	close (FH);
	
	if (!($$preferences{refresh_display})){
		$$preferences{refresh_display} = 60;
	}

	
	if (!($$preferences{server})){
		$$preferences{server}{"OpenVPN-Control"}{server_ip} = "127.0.0.1";
		$$preferences{server}{"OpenVPN-Control"}{port} = "7505";
		$$preferences{server}{"OpenVPN-Control"}{ssh_port_forward} = "7506";
		$$preferences{server}{"OpenVPN-Control"}{enabled} = 1;
	}
	
	
	if (!($$preferences{ssh})){
		if ($os eq "other"){
			$$preferences{ssh} = "/usr/bin/ssh";
		} else {
			$$preferences{ssh} = "putty.exe";
		}
	}
	
	#convert refresh into 1/4 seconds
	$refresh = ($$preferences{refresh_display} * 4);
	
	return ($preferences);


}

sub get_file {
	
	my $text;
	
	if ($os eq "windows"){

		$$preferences{ssh} = $config->getOpenFile(-title => "Where is your Putty.exe file",
											  -initialfile  => "putty.exe");	
											  
		$$preferences{ssh} =~ s/\\/\//g;									  
	
	} else {
		$$preferences{ssh} = $config->getOpenFile(-title => "Where is your ssh file, (usually /usr/bin/ssh)",
											  -initialdir => "/usr/bin",
											  -initialfile  => "ssh");
	}
	
	
	
}

sub Error_Message {
	
	my $message = shift;	
	
	my $message_dialog = $main->Dialog ( -title => "ERROR!!",
				              -buttons => [ "Ok" ],
					      	 -default_button => "Ok",
					      	 );

	$message_dialog->iconimage($icon);
	my $message_label = $message_dialog->Label(-text => $message,
	 				     )->pack(-expand => 1, -fill => 'x');

	
	#determine center of current window for placement
	my $x = $main->rootx;
	my $y = $main->rooty;
	$x = $x + 300;
	$y = $y + 100;

	#This could be more easily done,
	#but the window would not push to the forground under windows.
	$message_dialog-> Post($x, $y);
    $message_dialog-> grab;
    $message_dialog-> focusForce;
    $message_dialog-> Wait;
    
	return;
}

sub Info_Message {
	
	my $message = shift;	
	
	my $message_dialog = $main->Dialog ( -title => "OpenVPN Control",
				              -buttons => [ "Ok" ],
					      	 -default_button => "Ok",
					      	 );

	$message_dialog->iconimage($icon);
	my $message_label = $message_dialog->Label(-text => $message,
	 				     )->pack(-expand => 1, -fill => 'x');

	
	#determine center of current window for placement
	my $x = $main->rootx;
	my $y = $main->rooty;
	$x = $x + 300;
	$y = $y + 100;

	#This could be more easily done,
	#but the window would not push to the forground under windows.
	$message_dialog-> Post($x, $y);
    $message_dialog-> grab;
    $message_dialog-> focusForce;
    $message_dialog-> Wait;
    
	return;
	
}

sub confirm {
	
	my $message = shift;	
	
	my $message_dialog = $main->Dialog ( -title => "Confirm",
				              -buttons => [ "Yes", "No" ],
					      	 -default_button => "Yes",
					      	 );

	$message_dialog->iconimage($icon);
	my $message_label = $message_dialog->Label(-text => $message,
	 				     )->pack(-expand => 1, -fill => 'x');

	
	#determine center of current window for placement
	my $x = $main->rootx;
	my $y = $main->rooty;
	$x = $x + 300;
	$y = $y + 100;

	#This could be more easily done,
	#but the window would not push to the forground under windows.
	$message_dialog-> Post($x, $y);
    $message_dialog-> grab;
    $message_dialog-> focusForce;
    $message_dialog-> Wait;
    my $button = $message_dialog-> {selected_button};
    
	if ($button eq "Yes"){
		return;	
	} else {
		die;
	}

}

sub delete_server {
	
	my $server = shift;
	
	if (!($server_array[1])){
		Error_Message ("Can not remove last OpenVPN server");	
	}
	
	delete ($$preferences{server}{$server});
	
	@server_array = sort keys(%{$$preferences{server}});
	$server = $server_array[0];
	
	change_config_profile($server);
	return();
	
}

sub rename_server {

	my $server = shift;
	my $old_server = $server;
	
	my $new_server = $config->Dialog ( -title => "Rename Server $server",
				     				  -buttons => [ "Ok", "Cancel" ],
				     				  -default_button => "Ok");	
	
	$new_server->iconimage($icon);
	my $server_name_label = $new_server->Label(-text => "User Friendly name of OpenVPN Server: ",
	 				     )->pack(-expand => 1, -fill => 'x');

	my $server_address = $new_server->Entry (-background => "white",
					     -textvariable => \$server)->pack(-expand => 1, -fill => 'x');
					     
	#my $status = $new_server->Show(-popover => $config);			
	#determine center of current window for placement
	my $x = $config->rootx;
	my $y = $config->rooty;
	$x = $x + 300;
	$y = $y + 100;

	#display password prompt. This could be more easily done,
	#but the password entry field would not be the default focus.
	$new_server-> Post($x, $y);
    $server_address->focus;
    $new_server-> grab;
    $new_server-> focusForce;
    $new_server-> Wait;
    
     my $status = $new_server-> {selected_button};
    
	if ($status eq "Ok"){	
		
		$$preferences{server}{$server} = $$preferences{server}{$old_server};
		delete ($$preferences{server}{$old_server});
		
		@server_array = sort keys(%{$$preferences{server}});
	
		change_config_profile	($server);

		return();
	}
		
	return();
}

sub choose_ssh_port {
	
	my $port;
	my $old_port = 0;
	my $new_port;
	my $profile;
	
	foreach $profile (sort keys %{$$preferences{server}}){
		   
		   $port = $$preferences{server}{$profile}{ssh_port_forward};
		   if ($port > $old_port){
		   		$new_port = $port + 1;
		   }
		   $old_port = $port;

	}		
	
	foreach $profile (sort keys %{$$preferences{server}}){
		    $port = $$preferences{server}{$profile}{port};
		    if ($port == $new_port){
		   		$new_port = $port + 1;
		   }		
	}
	
	return ($new_port)
}

sub new_server {

	my $server;
	
	my $new_server = $config->Dialog ( -title => "Create New Server",
				     				  -buttons => [ "Ok", "Cancel" ]);	
	
	$new_server->iconimage($icon);
	my $server_name_label = $new_server->Label(-text => "User Friendly name of OpenVPN Server: ",
	 				     )->pack(-expand => 1, -fill => 'x');

	my $server_address = $new_server->Entry (-background => "white",
					     -textvariable => \$server)->pack(-expand => 1, -fill => 'x');
			     
	#determine center of current window for placement
	my $x = $config->rootx;
	my $y = $config->rooty;
	$x = $x + 300;
	$y = $y + 100;

	#display password prompt. This could be more easily done,
	#but the password entry field would not be the default focus.
	$new_server-> Post($x, $y);
    $server_address->focus;
    $new_server-> grab;
    $new_server-> focusForce;
    $new_server-> Wait;		
	
	 my $status = $new_server-> {selected_button};
	     
	if ($status eq "Ok"){	

		$$preferences{server}{$server}{server_ip} = "127.0.0.1";
		$$preferences{server}{$server}{port} = "7505";

	    	#choose ssh prort forwarding port
		$$preferences{server}{$server}{ssh_port_forward} =  choose_ssh_port();
		$$preferences{server}{$server}{enabled} = 1;

		change_config_profile($server);				
	
		write_preferences();
	}
	
	return;
}

sub ssh_config_defaults {

	my $server = shift;
	
	if ($$preferences{server}{$server}{ssh_tunnel} == 1){
		$$preferences{server}{$server}{server_ip} = "127.0.0.1";
		$$config_fields{textvariable}{server_ip}->configure(-state=>"disabled",
								  -background => 'gray');
								  
		$$config_fields{textvariable}{ssh_server}->configure(-state=>"normal",
							  -background => 'white');
							  
		$$config_fields{textvariable}{ssh_username}->configure(-state=>"normal",
							  -background => 'white');
							  
		
	} else {
		$$config_fields{textvariable}{server_ip}->configure(-state=>"normal",
							  -background => 'white');
							  
		$$config_fields{textvariable}{ssh_server}->configure(-state=>"disabled",
							  -background => 'gray');
							  
		$$config_fields{textvariable}{ssh_username}->configure(-state=>"disabled",
							  -background => 'gray');

	}

	$config->update();
	
	return();
	
	
}

sub configure_control {

	my $choice;
	my $button_label;
	my $ssh_server;
	my $ssh_username;
	my $ssh_port_forward;
	my $server_address;
	
	$preferences = get_preferences();
	
	if (!$server){
		print "Why is $server variable blank????\n";
		$server = $server_array[0];
	} 
		
	$config = $main->Dialog ( -title => "OpenVPN Control Configuration",
				     -buttons => [ "Ok", "Cancel" ]);

    $config->iconimage($icon);
	my $break1_label = $config->Label(-text => " ",
									-height => 2,
	 				     			)->pack(-expand => 1, -fill => 'x');
	 				     
	my $add_server_button = $config->Button( -text   => "Add Server",
               				   			   -command => [ sub { new_server(); }] 
               				   			   )->place(-x=>60, -y=>2);

	my $rename_server_button = $config->Button( -text   => "Rename Server",
               				   			   -command => [ sub { rename_server($server); }] 
               				   			   )->place(-x=>172, -y=>2);

	my $delete_server_button = $config->Button( -text   => "Delete Server",
               				   			   -command => [ sub { delete_server($server); }] 
               				   			   )->place(-x=>310, -y=>2);              				   			                 				   			   
	
	my $server_name_label = $config->Label(-text => "User Friendly name of OpenVPN Server: ",
	 				     )->pack(-expand => 1, -fill => 'x');
	
	 				     
	$server_popup = $config->BrowseEntry(    -autolistwidth => 1,
										   -textvariable => \$server, 
										   -background => "white",
										   -style => "MSWin32",
										   -choices => \@server_array,
										   -browsecmd => [ sub {$server = change_config_profile($server); } ]) 
										   ->pack();
	
	$$config_fields{variable}{enabled} = $config->Checkbutton(-text => "Enabled", 
							 					-variable => \$$preferences{server}{$server}{enabled},
							 					) ->pack(-expand => 1, -fill => 'x');
							 							     
	my $server_address_label = $config->Label(-text => "OpenVPN Server IP Address or DNS Name: ",
	 				     	  )->pack(-expand => 1, -fill => 'x');
	$$config_fields{textvariable}{server_ip} = $config->Entry (-background => "white",
					     -textvariable => \$$preferences{server}{$server}{server_ip})->pack(-expand => 1, -fill => 'x');
					     
	my $server_port_label = $config->Label(-text => "OpenVPN Management Server Port: ",
	 				     	  )->pack(-expand => 1, -fill => 'x');
	$$config_fields{textvariable}{port} = $config->Entry (-background => "white",
					     -textvariable => \$$preferences{server}{$server}{port})->pack(-expand => 1, -fill => 'x');

	my $refresh_time_label = $config->Label(-text => "How many seconds between screen refreshes: ",
	 				     	)->pack(-expand => 1, -fill => 'x');
	my $refresh_time = $config->Entry (-background => "white",
					     -textvariable => \$$preferences{refresh_display})->pack(-expand => 1, -fill => 'x');


	my $SSH_Section_Break = $config->Label(-text => "\n----------- SSH Port Forwarding Management Console------------\n",
	 				     		)->pack(-expand => 1, -fill => 'x');

	
	$$config_fields{variable}{ssh_tunnel} = $config->Checkbutton(-text => "Use SSH to create Tunnell to remote OpenVPN Management Server", 
							 					-variable => \$$preferences{server}{$server}{ssh_tunnel},
							 					-command => [ sub { ssh_config_defaults($server); }]
							 					) ->pack(-expand => 1, -fill => 'x');

	if ($os eq "windows"){
			my $ssh_label = $config->Label(-text => "Location of Putty.exe: ",
	 				     	)->pack(-expand => 1, -fill => 'x');
			
			$button_label = "Location Putty.exe: ";
			
	} else {
		
			$button_label = "Location of ssh file: ";
			my $ssh_label = $config->Label(-text => "Path to SSH Program ",
	 				     	)->pack(-expand => 1, -fill => 'x');
		
	}
	
	my $ssh= $config->Entry (-background => "white",
							-state => 'disabled',
					      	-textvariable => \$$preferences{ssh})->pack(-expand => 1, -fill => 'x');
	
	my $get_file = $config->Button( -text   => "$button_label",
								  -state => 'disabled',
               				      -command => \&get_file )->pack();
               	     
	my $ssh_server_label = $config->Label(-text => "IP Address or DNS Name to use for SSH Connection",
	 				     		)->pack(-expand => 1, -fill => 'x');
	 				     		
	$$config_fields{textvariable}{ssh_server} = $config->Entry (-background => "white",
						     -textvariable => \$$preferences{server}{$server}{ssh_server})->pack(-expand => 1, -fill => 'x');

	my $ssh_username_label = $config->Label(-text => "User Name for SSH login",
	 				     		)->pack(-expand => 1, -fill => 'x');
	$$config_fields{textvariable}{ssh_username} = $config->Entry (-background => "white",
						     -textvariable => \$$preferences{server}{$server}{ssh_username})->pack(-expand => 1, -fill => 'x');					     	
			
	my $ssh_port_forward_label = $config->Label(-text => "Available Local Port for Port Forwarding",
	 				     						)->pack(-expand => 1, -fill => 'x');
	
	$$config_fields{textvariable}{ssh_port_forward} = $config->Entry (-background => "grey",
						     		  -textvariable => \$$preferences{server}{$server}{ssh_port_forward},
									  -state => 'disabled',
									  -width => 8,
	 				     				)->pack();
	 				     				 	
	my $ssh_port_custom = $config->Checkbutton(-text => "Advanced SSH Settings.", 
							 					-variable => \$$preferences{server}{$server}{ssh_custom},
							 					-command => [ sub { if ($$preferences{server}{$server}{ssh_custom} == 1){
							 									$$config_fields{textvariable}{ssh_port_forward} ->configure(-state=>'normal',-background=>'white');
							 									$ssh->configure(-state=>'normal',-background=>'white');
							 									$get_file->configure(-state=>'normal',-background=>'white');
							 								} else {
							 									$$config_fields{textvariable}{ssh_port_forward} ->configure(-state=>'disabled',-background=>'gray');
							 									$ssh->configure(-state=>'disabled',-background=>'white');
							 									$get_file->configure(-state=>'disabled',-background=>'white');
							 								}
							 							} ]
							 					) ->pack();
	
	 				     				
	ssh_config_defaults($server);
	
	#my $status = $config->Show(-popover => $main);			
    	#determine center of current window for placement
	my $x = $main->rootx;
	my $y = $main->rooty;
	$x = $x + 100;


	#This could be more easily done,
	#but the window would not push to the forground under windows.
	$config-> Post($x, $y);
    $config-> grab;
    $config-> focusForce;
    $config-> Wait;
    my $status = $config-> {selected_button};
    
	if ($status eq "Ok"){	

		#ensure that no port forward conflisct exist
		prevent_port_conflicts();
		
		write_preferences();

		#refresh display
		get_login_info();
		
	}
	
	refresh();
	return;

}


sub change_config_profile {
	
	 $server = shift;

	 if (!($server)){
	 	
	 	#clear and re-create server array
	 	@server_array = undef;
		my $counter = 0;
	 	foreach (sort keys %{$$preferences{server}}){
	 		if ($_){
	 			$server_array[$counter] = $_;
	 			$counter++;
	 		}
	 	}
	 	
	 	#set server to first choice
	 	$server = $server_array[0];
	 }
	 
	 my $profile = $server;
	 
	 my $field;
	 my $type;
	 
	 foreach $type (sort keys %{$config_fields}){
	 	foreach $field (sort keys %{$$config_fields{$type}}){
	 		if ($type eq "variable"){
	 			$$config_fields{$type}{$field}->configure(-variable => \$$preferences{server}{$server}{$field});
	 		} elsif ($type eq "textvariable"){
	 			$$config_fields{$type}{$field}->configure(-textvariable => \$$preferences{server}{$server}{$field});
	 		}
	 	}
	 }
	 
	 ssh_config_defaults($server);
	 
	 	 
	 #rebuild server list
	 @server_array = undef;
	 my $counter = 0;
	 foreach (sort keys %{$$preferences{server}}){
	 	if ($_){
	 		$server_array[$counter] = $_;
	 		$counter++;
	 	}
	 }
	 
	 	 
	 $server_popup->configure(-choices => \@server_array);
	 $server_popup->update();
	 
	 $server_popup->configure(-textvariable=>\$server);
	 $config->update();

	 $server = $profile;
	 
	 #print "Server is equal to $server\n";
	 return($server);
	
}

sub prevent_port_conflicts {
	
	my $port;
	my $server;
	my $status = "ok";
	
	my $server1;
	my $server2;
	
	foreach $server (sort keys %{$$preferences{server}}){
		
		if ($$preferences{server}{$server}{ssh_tunnel} == 1){
			foreach (sort keys %{$$preferences{server}}){
				 if ($$preferences{server}{$_}{ssh_port} == 1){
				 	if ($$preferences{server}{$server}{ssh_port_forward} eq $$preferences{server}{$_}{ssh_port_forward}){
				 			if ($server ne $_){
				 				if (($$preferences{server}{$_}{enabled} == 1) & ($$preferences{server}{$server}{enabled} == 1)){
									$status = "error";
									$server1 = $server;
									$server2 = $_;
				 				}
				 			}
				 	}
				 } else {
					if ($$preferences{server}{$server}{ssh_port_forward} eq $$preferences{server}{$_}{port}){
				 			if ($server ne $_){
				 				if (($$preferences{server}{$_}{enabled} == 1) & ($$preferences{server}{$server}{enabled} == 1)){
									$status = "error";
									$server1 = $server;
									$server2 = $_;
				 				}
				 			}
					}
				 }
			}
							 		
		} else {
			foreach (sort keys %{$$preferences{server}}){
				 if ($$preferences{server}{$_}{ssh_tunnel} == 1){
				 	if ($$preferences{server}{$server}{port} eq $$preferences{server}{$_}{ssh_port_forward}){
				 			if ($server ne $_){
				 				if (($$preferences{server}{$_}{enabled} == 1) & ($$preferences{server}{$server}{enabled} == 1)){
									$status = "error";
									$server1 = $server;
									$server2 = $_;
				 				}
				 			}
				 	}
				 } else {
					if ($$preferences{server}{$server}{port} eq $$preferences{port}){
					 		if ($server ne $_){
				 				if (($$preferences{server}{$_}{enabled} == 1) & ($$preferences{server}{$server}{enabled} == 1)){
									$status = "error";
									$server1 = $server;
									$server2 = $_;
				 				}
				 			}
					}
				 }
			}
			
		}
	}	
	
	if ($status eq "error"){
		$$preferences{server}{$server1}{enabled} = 0;
		$$preferences{server}{$server2}{enabled} = 0;
		Error_Message("Port conflicts between server\n$server1 and $server2!!\nBoth servers have been disabled.
Go back into configure to resolve.");
	}
	return;		
}

sub confirm_ssh_link {

	my $confirm = "no";
	my $server = shift;
	my $count = 0;
	
	my $port;
	
	if ($$preferences{server}{$server}{ssh_port_forward}){
		$port = $$preferences{server}{$server}{ssh_port_forward};
	} else {
		$port = $$preferences{server}{$server}{port};
	}
	
	my $timeout = 15;
	
	while ($confirm eq "no"){
		$confirm = "yes";
			
		$connection = IO::Socket::INET->new( PeerAddr => $$preferences{server}{$server}{server_ip},
				     PeerPort => $port,
				     Proto => "tcp",
				     Type => SOCK_STREAM) or $confirm = "no";
					     
		
		sleep(2);
		$count ++;
		if ($count == $timeout){  #the program has been waiting for at least 60 seconds with no connection
		
			Error_Message ("Timing out.\nCould not establish SSH connection with $server $$preferences{server}{$server}{ssh_server}");
			$$preferences{server}{$server}{enabled} = 0;
			return;
		}
	}
	
	close ($connection);
	
	Info_Message ("SSH Tunnel has been established to $server");
		     
	return;
}

sub check_ssh {
		
		my $server = shift;
		my $status = "ok";
		
		my $connection = IO::Socket::INET->new( PeerAddr => $$preferences{server}{$server}{ssh_server},
				     PeerPort => '22',
				     Proto => "tcp",
				     Type => SOCK_STREAM) or $status = "error";
		
		if ($status eq "ok"){
			close ($connection);
		
		} else {    
		  	Error_Message ("Could not communicate with $$preferences{server}{$server}{ssh_server}");
		  	$$preferences{server}{$server}{ssh_tunnel} = 0;
		}

		return($status);
	
}

sub ssh_tunnel {
	
	my $server = shift;
	my $status;

	
	if ($$preferences{server}{$server}{ssh_tunnel} == 1){
				
		$initial_ssh_session = "yes";
		
		$status = check_ssh($server);
		if ($status eq "error"){
			return();
		}
		
		#write server to launch
		open (FH, "> server.tmp");
		print FH $server;
		close(FH);
		
		$ssh_thread = Thread->new(\&thread_ssh);

		confirm_ssh_link($server);
	
	}
		
	return;

}

sub thread_ssh {
			

	my $os;
	if ($ENV{'windir'}){
		$os = "windows";
	} else {
		$os = "other";
	}

	my $preferences = get_preferences();
	
	open (FH, "< server.tmp");
	$server = <FH>;
	close(FH);
	unlink ("server.tmp");
		
	my $dir;
	
	if ($$preferences{ssh} eq "putty.exe"){
	 	$dir = getcwd . "/";
	} else {
		$dir = "";
	}
	
	if ($os eq "windows"){

		 my $proc_id;
		 my $procHandle;
		 
		 #Win32::Process::Create($procHandle,
		 #						"${dir}$$preferences{ssh}",
		 #						"$$preferences{ssh} -ssh -N -L $$preferences{server}{$server}{ssh_port_forward}:127.0.0.1:$$preferences{server}{$server}{port} -l $$preferences{server}{$server}{ssh_username} $$preferences{server}{$server}{ssh_server}",
		 #						'0',
		 #						CREATE_SEPARATE_WOW_VDM,
		 #						'.');
		 						
		 $proc_id = $procHandle->GetProcessID();
		 open (FH, "> ${server}-proc");
		 print FH $proc_id;
		 close (FH);
		 	
	} else {

		`xterm -title "OpenVPN_Control_$server" -geometry 65x10 -e $$preferences{ssh} -N -L $$preferences{server}{$server}{ssh_port_forward}:127.0.0.1:$$preferences{server}{$server}{port} -l $$preferences{server}{$server}{ssh_username} $$preferences{server}{$server}{ssh_server}`;

	}
			
}

sub write_preferences {

	my $conf_file = "openvpn_control.conf";
	my $key;
	my $status = "ok";
	my $server;
	
	$refresh = ($$preferences{refresh} * 4);
	
	#confirm ssh program exists
	if ($$preferences{server}{$server}{ssh_tunnel} == 1){
		if (!( -f $$preferences{ssh})){
			$preferences = get_preferences;
			if ($os eq "windows"){
					Error_Message ("Invalid path to putty.exe\nPreferences not saved"); 
			} else {
					Error_Message ("Invalid path to ssh, Preferences not saved");
			}
			return;
		}
	}
	
	open (FH, "> $conf_file") or $status = "error";
	
	if ($status eq "error"){
		Error_Message("Can not write to $conf_file");
		return;
	}
	
	foreach $key (sort keys %{$preferences}){
		if ($key eq "server"){
			foreach $server (@server_array){
				foreach (sort keys %{$$preferences{server}{$server}}){
					print FH "$server:$_=$$preferences{server}{$server}{$_}\n";
				}
			}
		} else {
			print FH "$key=$$preferences{$key}\n";
		}
	}

	close(FH);
	
	#convert refresh from seconds to 1/4 seconds
	$refresh = ($$preferences{refresh_display} * 4);
	
	return;
}

sub exit_handler {

	#my $ssh_pid = read_ssh_pid();
	
	my $ssh_tunnel = "no";
	my $process_list;
	my $process;
	my @processes;
	my @pids;
	my @window_names;

	foreach $server (keys %{$$preferences{server}}){
		if ($$preferences{server}{$server}{ssh_tunnel} == 1){
				$ssh_tunnel = "yes";
				push (@window_names, "OpenVPN_Control_$server");
		}
	}
	
	my $proc_id;
	if ($ssh_tunnel eq "yes"){

		if ($os eq "windows"){
			#Info_Message("You should now manually exit out of the Putty session that was launched!");
			foreach $server (keys %{$$preferences{server}}){
				open (FH, "< ${server}-proc");
		 		$proc_id =  <FH>;
		 		close (FH);
				Win32::Process::KillProcess($proc_id, 0);
				unlink ("${server}-proc");
			}
			
		} else {

			$process_list = `ps -ax`;
			@processes = split (/\n/, $process_list);
		
			foreach $process(@processes){
				foreach $server (@window_names){
					if ($process =~ /$server/){
						$process =~ /^([\s\d]+)\s.*/;
						kill 9 => $1;
					}
				}		
			}
		}
	}
	
	kill 9 => $$;
	exit;
	
} 

sub get_password {
	
	my $password;
	my $passwd = "";
	my $server = shift;
	
	my $password_dialog = $main->Dialog ( -title => "OpenVPN Control",
				              -buttons => [ "Ok", "Quit" ],
					      	 -default_button => "Ok",
					      	 );

	$password_dialog->iconimage($icon);
	my $password_label = $password_dialog->Label(-text => "OpenVPN $server Password",
	 				     )->pack(-expand => 1, -fill => 'x');
	$password = $password_dialog->Entry (-background => "white",
					-textvariable => \$passwd,
					-show => '*' )->pack(-expand => 1, -fill => 'x');

	
	#determine center of current window for placement
	my $x = $main->rootx;
	my $y = $main->rooty;
	$x = $x + 300;
	$y = $y + 100;

	#display password prompt. This could be more easily done,
	#but the password entry field would not be the default focus.
	$password_dialog-> Post($x, $y);
    $password->focus;
    $password_dialog-> grab;
    $password_dialog-> focusForce;
    $password_dialog-> Wait;
    my $button = $password_dialog-> {selected_button};

    
	if ($button eq "Quit"){
		exit_handler();
	}
	
	return($passwd);

}

sub about {

	my $message = "\nOpenVPN Control 1.02e\n\nThis is a management console for the product OpenVPN\n\n";
	$message .= "Copyright 2005, Ron Dorn\nThis product is being distributed under the GPL license\n\n";
	$message .= "http://sourceforge.net/projects/openvpn-control/\n";
	
	Info_Message($message);


}

sub fix_time {
	
	my $in_time = shift;
	
	my $ampm;
	
	$in_time =~ /(\d+):(\d+)/;
	
	my $hour = $1;	
	my $min = $2;
	
	if ($hour == 0){
		$hour = "12";
		$ampm = "am";
	
	} elsif ($hour >= 13){
		$hour = $hour - 12;
		$ampm = "pm";
		
	} elsif ($hour == 12){
		$ampm = "pm";
	
	} else {
		$ampm = "am";
	}
	
	my $out_time = "${hour}:${min} $ampm";
	
	return ($out_time);
	
}